///////////////////////////////////////////////////////////////////////////////
// NABLA - a Numerical Analysis Based LAnguage                               //
//                                                                           //
// Copyright (C) 2014~2016 CEA/DAM/DIF                                       //
// IDDN.FR.001.520002.000.S.P.2014.000.10500                                 //
//                                                                           //
// Contributor(s): CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr          //
//                                                                           //
// This software is a computer program whose purpose is to translate         //
// numerical-analysis specific sources and to generate optimized code        //
// for different targets and architectures.                                  //
//                                                                           //
// This software is governed by the CeCILL license under French law and      //
// abiding by the rules of distribution of free software. You can  use,      //
// modify and/or redistribute the software under the terms of the CeCILL     //
// license as circulated by CEA, CNRS and INRIA at the following URL:        //
// "http://www.cecill.info".                                                 //
//                                                                           //
// The CeCILL is a free software license, explicitly compatible with         //
// the GNU GPL.                                                              //
//                                                                           //
// As a counterpart to the access to the source code and rights to copy,     //
// modify and redistribute granted by the license, users are provided only   //
// with a limited warranty and the software's author, the holder of the      //
// economic rights, and the successive licensors have only limited liability.//
//                                                                           //
// In this respect, the user's attention is drawn to the risks associated    //
// with loading, using, modifying and/or developing or reproducing the       //
// software by the user in light of its specific status of free software,    //
// that may mean that it is complicated to manipulate, and that also         //
// therefore means that it is reserved for developers and experienced        //
// professionals having in-depth computer knowledge. Users are therefore     //
// encouraged to load and test the software's suitability as regards their   //
// requirements in conditions enabling the security of their systems and/or  //
// data to be ensured and, more generally, to use and operate it in the      //
// same conditions as regards security.                                      //
//                                                                           //
// The fact that you are presently reading this means that you have had      //
// knowledge of the CeCILL license and that you accept its terms.            //
//                                                                           //
// See the LICENSE file for details.                                         //
///////////////////////////////////////////////////////////////////////////////
with ℝ²;
#include "assert.h"
ofstream output("evol.plot");
ofstream output2("mesh.plot");
ofstream output3("output.plot");
ofstream results("result.plot");

// ****************************************************************************
// * Options du schéma
// ****************************************************************************
options{
  // Mesh options
  ℝ LENGTH                    = 1.0;
  ℕ X_EDGE_ELEMS              = 8;
  ℕ Y_EDGE_ELEMS              = 8;
  ℕ Z_EDGE_ELEMS              = 1;

  Bool DEBUG = false;
  Bool option_test_sod = true;
  Bool option_semi_implicit = false;
  Integer option_case = 2;
  Real option_coupling_coef=1;
  Real option_ini_zg_p = 1.0;
  Real option_ini_zd_p = 1.0;
  Real option_ini_zg_u = 0.0;
  Real option_ini_zd_u = 0.0;
  Real option_ini_zg_t = 1.0;
  Real option_ini_zd_t = 1.0;

  Real option_x_min = 0.0;
  Real option_y_min = 0.0;
  Real option_x_interface = 0.5;
  Real option_x_max = LENGTH;
  Real option_y_max = LENGTH;
  
  Real option_δt_initial = 1e-7;         // ! utilisé en dur dans le backend lambda !
  Real option_δt_min = 1e-15;
  Real option_δt_max = 1.e-1;
  Real option_stoptime = 1.;             // ! utilisé en dur dans le backend lambda !
  Real option_δt_control = 0.15;           // Contrôle du pas de temps hydrodynamique

  Real option_δt_min_variation = 0.99;     // Décroissance maximale tolérée du pas de temps
  Real option_δt_max_variation = 0.10;     // Amplitude maximale de croissance du pas de temps
  Bool option_δt_last_exact = true;
  Bool option_ini_exact = false;
  
  Bool option_dump_at_end =true;
  Bool option_dump_output = true;          
  Integer option_max_iterations = 0;   // ! utilisé en dur dans le backend lambda !
  Integer dump_iterations = 1000;   
  Real option_rdq_α=0.4;
};

// ****************************************************************************
// * Déclaration des variables aux noeuds
// ****************************************************************************
nodes{
  Real3   node_u;                          // Vitesse nodale du maillage
  Real3x3 node_u_matrix;                   // Matrice aux sommets pour le calcul de la vitesse
  Real3   node_u_second_member;            // Second membre aux sommets pour le calcul de la vitesse
};

// ****************************************************************************
// * Déclaration des variables aux mailles
// ****************************************************************************
cells{  
  Real3 center;
  Real V;                                  // Volume des mailles
  Real p;                                  // Pression
  Real3 u;                                 // Vitesse moyenne
  Real temperature;                        // Temperature
  Real pana;                               // Pression analytique
  Real3 uana;                              // Vitesse moyenne analytique
  Real tana;                               // Temperature analytique
  Real p1_δt;
  Real absCQs[nodes];
  Real3 CQs[nodes];                        // Résultantes aux sommets
  Real3x3 AQs[nodes];                      // Matrices de contributions aux sommets
  Real3x3 BQs[nodes];                      // Matrices de contributions aux sommets
  Real3 reconstructed_u[nodes];            // Tableau des vitesses reconstruites
  Real reconstructed_p[nodes];             // Tableau des pressions reconstruites
  Real3 velocity_fluxes_Σ;
  Real pressure_fluxes_Σ;
  Real temperature_source_Σ;
  Real minCellDelta;
};

// ****************************************************************************
// * Globals
// ****************************************************************************
global{
  ℝ global_δt;
  ℝ e_tp;
  ℝ Δl;
  ℝ minΔ;
  ℝ sigma_a;
  ℝ sigma_s;
  ℝ sigma;
  ℝ epsilon;
};

// ****************************************************************************
// * Partie boucle de calcul ]+0.0,+∞[ 
// ****************************************************************************

// ****************************************************************************
// * computeAQsMatrices
// ****************************************************************************
∀ cells void computeAQsMatrices(void)@ 9.0 {
  const Real lambda = 1.0;
  ∀ node{
    AQs = CQs⊗CQs;
    AQs *= lambda/absCQs;
    if (DEBUG)
      printf("\n[31m[std] AQs[%d;node #%d]=\n\t\t[\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]][0m",uid,n,
           AQs.x.x,AQs.x.y,AQs.x.z,
           AQs.y.x,AQs.y.y,AQs.y.z,
           AQs.z.x,AQs.z.y,AQs.z.z);
  }
}

// ****************************************************************************
// * computeBQsMatrices
// ****************************************************************************
∀ cells void computeBQsMatrices(void)@ 9.1 {
  const Real3 ccell=center;
  const Real3 X[4] = {coord[0],coord[1],coord[2],coord[3]};

  ∀ node{
    const int i=#;
    BQs = CQs⊗(X[i]-ccell);
    BQs *= sigma_s/epsilon;
    if (DEBUG)
      printf("\n\t[31m[std] BQs[%d]=\n\t\t[\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]][0m",uid,
           BQs.x.x,BQs.x.y,BQs.x.z,
           BQs.y.x,BQs.y.y,BQs.y.z,
           BQs.z.x,BQs.z.y,BQs.z.z);
  }
}

// ****************************************************************************
// * temporalSchemeReconstructPressure
// ****************************************************************************
∀ cells @ 10.0{
  ∀ node{
    reconstructed_p=p;
    if (DEBUG)
      printf("\n\t[32m[std] reconstructed_p=%.15e[0m",
             reconstructed_p);
  }
}

// ****************************************************************************
// * temporalSchemeReconstructVelocity
// ****************************************************************************
∀ cells void temporalSchemeReconstructVelocity(void)@ 10.0{
  ∀ node{
    reconstructed_u=u;
    if (DEBUG)
      printf("\n\t[32m[std] reconstructed_u=(%.15e,%.15e,%.15e)[0m",
           reconstructed_u.x,reconstructed_u.y,reconstructed_u.z);
  }
}

// ****************************************************************************
// * temporalSchemeIniMatricesAndSecondMember
// ****************************************************************************
∀ nodes void temporalSchemeIniMatricesAndSecondMember(void)@ 11.0{
  node_u_matrix=Real3x3();
  node_u_matrix.z=Real3(0,0,1);
  node_u_second_member=0;
}

// ****************************************************************************
// * temporalSchemeAssembleMatricesAndSecondMember
// ****************************************************************************
∀ cells void temporalSchemeAssembleMatricesAndSecondMember(void)@ 12.0{
  ∀ node{
    const Real3x3 local_matrix = AQs;
    node_u_matrix += AQs;
    node_u_matrix += BQs;
    node_u_second_member += local_matrix ⨂ reconstructed_u;
    node_u_second_member += CQs*reconstructed_p;
  }
}
∀ nodes @ 12.1{ // On revient APRES les écritures aux noeuds pour dumper
  if (DEBUG)
    printf("\n\t[31m[std] Before BOUNDARIES node_u_matrix[%d]=\n\t\t[\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]][0m",uid,
           node_u_matrix.x.x,node_u_matrix.x.y,node_u_matrix.x.z,
           node_u_matrix.y.x,node_u_matrix.y.y,node_u_matrix.y.z,
           node_u_matrix.z.x,node_u_matrix.z.y,node_u_matrix.z.z);
  if (DEBUG)
    printf("\n\t[31m[std] Before BOUNDARIES node_u_second_member[%d]=\
(%.15e,%.15e,%.15e)[0m",uid,
           node_u_second_member.x,
           node_u_second_member.y,
           node_u_second_member.z);
}

// ****************************************************************************
// * Le calcul aux boundaries se fait ici @ 13.1,13.2
// ****************************************************************************

// ****************************************************************************
// * synchronize pour Arcane node_u_matrix & node_u_matrix_second_member
// ****************************************************************************
void synchronizeNodeUMatrixAndSecondMember(void)@ 13.4{
  synchronize(node_u_matrix);
  synchronize(node_u_second_member);
}

// ****************************************************************************
// * temporalSchemeComputeNodesVelocity
// ****************************************************************************
∀ nodes void temporalSchemeComputeNodesVelocity(void)@ 14.0{
  if (DEBUG)
    printf("\n\t[31m[std] node_u_matrix=[0m\n\t\t[31m[\
[%.15e %.15e %.15e]\n\t\t[%.15e %.15e %.15e]\n\t\t[%.15e %.15e %.15e]][0m",
         node_u_matrix.x.x,node_u_matrix.x.y,node_u_matrix.x.z,
         node_u_matrix.y.x,node_u_matrix.y.y,node_u_matrix.y.z,
         node_u_matrix.z.x,node_u_matrix.z.y,node_u_matrix.z.z);
  const Real determinent = matrixDeterminant(node_u_matrix);
  if (DEBUG)
    printf("\n\t[31m[std] node_u_second_member=(%.15e,%.15e,%.15e)[0m",
         node_u_second_member.x,
         node_u_second_member.y,
         node_u_second_member.z);
  assert(determinent!=0.0);
  node_u_matrix = inverseMatrix(node_u_matrix, determinent);
  node_u = node_u_matrix ⨂ node_u_second_member;
  node_u.z=0.0;
  if (DEBUG)
    printf("\n\t[31m[std] node_u[%d]=(%.15e,%.15e,%.15e)[0m",
           uid,node_u.x,node_u.y,node_u.z);
}
  
// ****************************************************************************
// * temporalSchemeIniFluxesSum
// ****************************************************************************
∀ cells void temporalSchemeIniFluxesSum(void)@ 15.0{
  velocity_fluxes_Σ = Real3(0.,0.,0.);
  pressure_fluxes_Σ = 0.0;
  temperature_source_Σ = 0.0;
}

// ****************************************************************************
// * temporalSchemeComputeStandardFluxesSum
// ****************************************************************************
∀ cells void temporalSchemeComputeStandardFluxesSum(void) @ 16.0{
    
//  temperature_source_Σ -= e_tp * (temperature - p);    
  ∀ node{
    const Real3 Δu = -node_u+reconstructed_u;
    if (DEBUG) printf("\n\t[32m[n+1] Δu[%d]=(%.15e,%.15e)[0m",uid,Δu.x,Δu.y);   
//    Real3 FQs = (CQs⋅Δu)/absCQs*CQs;
//    if (DEBUG) printf("\n\t[32m[n+1] 1:FQs[%d]=(%.15e,%.15e)[0m",uid,FQs.x,FQs.y);
    Real3 FQs = AQs ⨂ Δu;
    if (DEBUG) printf("\n\t[32m[n+1] 2:FQs[%d]=(%.15e,%.15e)[0m",uid,FQs.x,FQs.y);
//    FQs += reconstructed_p*CQs;
    velocity_fluxes_Σ -= FQs/epsilon;
    pressure_fluxes_Σ -= (CQs⋅node_u)/epsilon;
  }
  if (DEBUG) printf("\n\t[32m[n+1] pressure_fluxes[%d]=(%.15e)[0m",uid,pressure_fluxes_Σ);  
}

// ****************************************************************************
// * temporalSchemeApplyFluxesStandard
// ****************************************************************************
∀ cells void temporalSchemeApplyFluxesStandard(void)@ 19.0{
  if (DEBUG) printf("\n\t[32m[n+1] deltaT=%.15e[0m",δt);
  if (DEBUG) printf("\n\t[32m[n+1] pre u[%d]=(%.15e,%.15e,%.15e)[0m",uid,u.x,u.y,u.z);
  if (DEBUG)
    printf("\n\t[32m[n+1] velocity_fluxes_Σ[%d]=(%.15e,%.15e,%.15e)[0m",uid,velocity_fluxes_Σ.x,velocity_fluxes_Σ.y,velocity_fluxes_Σ.z);
  u += δt*velocity_fluxes_Σ/V;
  if (DEBUG)
    printf("\n\t[32m[n+1] post u[%d]=(%.15e,%.15e,%.15e)[0m",uid,u.x,u.y,u.z);
  if (DEBUG) printf("\n\t[32m[n+1] pre p[%d]=(%.15e)[0m",uid,p);
  if (DEBUG)
    printf("\n\t[32m[n+1] pressure_fluxes_Σ[%d]=(%.15e)[0m",uid,pressure_fluxes_Σ);
  p += δt*pressure_fluxes_Σ/V;
  if (option_coupling_coef==0)
    e_tp = (sigma_a * δt)/(epsilon*epsilon+2.*sigma_a*δt);
  if (option_coupling_coef==1)
    e_tp = 0.5*(1.-exp(-2.*sigma_a*δt/(epsilon*epsilon)));
  if (option_coupling_coef==2)
    e_tp = sigma_a*δt/(epsilon*epsilon);
  if (option_semi_implicit){    
    temperature_source_Σ = e_tp/(1. + 2.*e_tp) * (p - temperature);
    p-= temperature_source_Σ;
    temperature += temperature_source_Σ;///V;
  } else {
    temperature_source_Σ = e_tp * (p - temperature);
    p-= temperature_source_Σ;
    temperature += temperature_source_Σ;///V;
  }
//  temperature_source_Σ -= e_tp * (temperature - p);    
  //p += -temperature_source_Σ;///V;

  if (DEBUG) printf("\n\t[32m[n+1] post p[%d]=(%.15e)[0m",uid,p);
}

// ****************************************************************************
// * geomComputeCQsQuad @ 23.0
// * geomComputeSurfaceAndVolume @ 24.0 
// ****************************************************************************

// ****************************************************************************
// * glaceEOS vient se placer ici: [26.0 → 30.0]
// ****************************************************************************

// ****************************************************************************
// * Réduction min vers le global_δt
// ****************************************************************************
∀ cells global_δt <?= p1_δt @ 31.0;

∀ cells void dumpGnuplot(void) @ 34.0 if (option_dump_at_end and (((iteration-1)%dump_iterations)==0)){
  if (uid==0) info()<<"\33[1;31m[dumpGnuplot] Now dumping to output.plot file!\33[0m";
  output3 << p << " ";
  ∀ node
    output3 << coord.x << " " << coord.y << " ";
  output3 << coord[0].x << " " << coord[0].y << " ";
  output3 << "\n";
  if (uid==NABLA_NB_CELLS-1) output3 << "\n\n";
}


// ****************************************************************************
// * Dump Output
// ****************************************************************************
∀ cells void dumpOutput(void) @ 35.0 if (((iteration-1)%dump_iterations)==0) {
  if (uid==0) info()<<"\33[1;31m[dumpOutput] Now dumping to evol.plot file!\33[0m";
  output  << "\t" << center.x 
          << "\t" << center.y
          << "\t" << p
          << "\t" << u.x
          << "\t" << u.y
          << "\t" << temperature
          << "\t" << pana
          << "\t" << uana.x
          << "\t" << uana.y
          << "\t" << tana
          << "\n";
  if (c==(NABLA_NB_CELLS-1)) output << "\n";
} 


// ****************************************************************************
// * glaceTimeStepCtrl
// ****************************************************************************
void hydroTimeStepCtrl(void) @ 35.0{
  Real new_δt=option_δt_control*global_δt;
  if (true)
    printf("\n\t[1m iteration=%d, time=%f, global_δt=%.15e[0m",iteration,time,global_δt);
  // on calcule les variations min et max autorisées
  const Real max_time_step = (1.0+option_δt_max_variation)*δt;
  const Real min_time_step = (1.0-option_δt_min_variation)*δt;
  if (new_δt > max_time_step) new_δt=max_time_step;
  if (new_δt < min_time_step) new_δt=min_time_step;
  new_δt = min(new_δt, option_δt_max);
  new_δt = max(new_δt, option_δt_min);
  // le dernier calcul se fait exactement au temps final
  if (option_δt_last_exact && ((time + new_δt) > option_stoptime)) 
      new_δt = option_stoptime - time;
  δt=new_δt;
}


// ****************************************************************************
// * Dump Solution
// ****************************************************************************
∀ cells void dumpSolution(void) @ 36.0 if (option_dump_output and (time>=(option_stoptime-δt))){ 
  if (uid==0) info()<<"\33[1;31m[dumpSolution] Now dumping to result.plot file!\33[0m";
  results << "\t" << center.x 
          << "\t" << center.y
          << "\t" << p
          << "\t" << u.x
          << "\t" << u.y
          << "\t" << temperature
          << "\t" << pana
          << "\t" << uana.x
          << "\t" << uana.y
          << "\t" << tana
          << "\n";
}


// ****************************************************************************
// * Test for QUIT
// ****************************************************************************
void testForQuit(void) @ 40.0{
  if (option_max_iterations!=0 && iteration >= option_max_iterations) exit;
  if ((time<(option_stoptime-δt))) return;
  printf("\n[1;33m final δt=%.15e[0m\n",δt);
  exit;
}
