///////////////////////////////////////////////////////////////////////////////
// NABLA - a Numerical Analysis Based LAnguage                               //
//                                                                           //
// Copyright (C) 2014~2016 CEA/DAM/DIF                                       //
// IDDN.FR.001.520002.000.S.P.2014.000.10500                                 //
//                                                                           //
// Contributor(s): TRAN Guillaume - Guillaume.Tran@polytechnique.edu         //
//                 LOISEAU Pascal - Pascal.Loiseau@cea.fr                    //
//                 CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr          //
//                                                                           //
// This software is a computer program whose purpose is to translate         //
// numerical-analysis specific sources and to generate optimized code        //
// for different targets and architectures.                                  //
//                                                                           //
// This software is governed by the CeCILL license under French law and      //
// abiding by the rules of distribution of free software. You can  use,      //
// modify and/or redistribute the software under the terms of the CeCILL     //
// license as circulated by CEA, CNRS and INRIA at the following URL:        //
// "http://www.cecill.info".                                                 //
//                                                                           //
// The CeCILL is a free software license, explicitly compatible with         //
// the GNU GPL.                                                              //
//                                                                           //
// As a counterpart to the access to the source code and rights to copy,     //
// modify and redistribute granted by the license, users are provided only   //
// with a limited warranty and the software's author, the holder of the      //
// economic rights, and the successive licensors have only limited liability.//
//                                                                           //
// In this respect, the user's attention is drawn to the risks associated    //
// with loading, using, modifying and/or developing or reproducing the       //
// software by the user in light of its specific status of free software,    //
// that may mean that it is complicated to manipulate, and that also         //
// therefore means that it is reserved for developers and experienced        //
// professionals having in-depth computer knowledge. Users are therefore     //
// encouraged to load and test the software's suitability as regards their   //
// requirements in conditions enabling the security of their systems and/or  //
// data to be ensured and, more generally, to use and operate it in the      //
// same conditions as regards security.                                      //
//                                                                           //
// The fact that you are presently reading this means that you have had      //
// knowledge of the CeCILL license and that you accept its terms.            //
//                                                                           //
// See the LICENSE file for details.                                         //
///////////////////////////////////////////////////////////////////////////////
with ℝ, ℵ;


// ****************************************************************************
// * αβγδεζηθικλμνξοπρςστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ
// * ∀ ℂℕℝℤℾ √∛ ½⅓¼⅛ 𝜕ℵ∂ * ⋅⨯⤫⊗⨂⊛ ∞ ⁰¹²³ⁱⁿ⁺⁼
// ****************************************************************************
//ofstream output("onde.plot");

options{
  ℕ nbmailles = 1024;
  // Mesh options
  ℝ LENGTH                    = 1.0;
  ℕ X_EDGE_ELEMS              = nbmailles;
  ℕ Y_EDGE_ELEMS              = 1;
  ℕ Z_EDGE_ELEMS              = 1;

  ℾ DEBUG = false;
  ℝ CFL = 2.0;
  ℝ option_CL = 7.0;

  ℾ hypre = true;
  ℾ tridiag = false;
    
  ℝ ε = 1.0e-6;  
  ℝ option_xmin = 0.0;
  ℝ option_xmax = 1.0;
  ℝ pi =atan(1.0)*4.0;
  
  // Yet required options for Okina 
  ℝ option_δt_initial = 0.001;
  ℝ option_stoptime=0.2;
  ℕ option_max_iterations = 100;   // ! utilisé en dur dans les backends

  // Options d'Aleph
  ℝ alephEpsilon             = 1.e-16;
  ℕ alephUnderlyingSolver    = 2;
  ℕ alephMaxIterations       = 2048;
  // DIAGONAL=0, AINV=1, AMG=2, IC=3, POLY=4, ILU=5, ILUp=6,
  // SPAIstat=7, SPAIdyn=8, DDMCriteriaAdaptedSelector=9, NONE=10
  ℕ alephPreconditionerMethod = 0;
  // PCG=0, BiCGStab=1 , BiCGStab2=2, GMRES=3, SAMG=4, QMR=5, SuperLU=6
  ℕ alephSolverMethod     = 3; 
  ℕ alephNumberOfCores    = 0;
  ℾ option_aleph_dump_matrix = false;
};


// ****************************************************************************
// * Cell Variables
// ****************************************************************************
cells{
  ℝ E;
  ℝ sousdiag,diag,surdiag;
  ℝ solution,temp;
};


// ****************************************************************************
// * Global Variables
// ****************************************************************************
global{
  ℝ last;
};


// ****************************************************************************
// * if (tridiag) → ResolutionSystemeTridiag
// ****************************************************************************
void solveTriDiag(void) @ 2.1{
  if (tridiag)
    ResolutionSystemeTridiag(sousdiag,diag,surdiag,
                             solution,temp,nbmailles);
}
∀ cells @ 2.2 if (tridiag) {E=solution;}


// ****************************************************************************
// * 
// ****************************************************************************
void tstForQuit(void) @ 4.0 {
  if ((time<(option_stoptime-δt))) return;
  exit;
}
